#include "global.h"
#include "RageLog.h"
#include "RageUtil.h"
#include "RageUtil_FileDB.h"
#include "test_misc.h"

class TestFilenameDB: public FilenameDB
{
public:
	/* Each "directory" contains 100 "files" numbered 00 through 99, and 10 "subdirectories"
	 * names "a" through "j". */
        void PopulateFileSet( FileSet &fs, const CString &sPath )
	{
		for( int i = 0; i < 100; ++i )
		{
			File f;
			f.SetName( ssprintf( "%02i", i ) );
			f.dir = false;
			f.size = 100;
			f.hash = 0;

			fs.files.insert(f);
		}

		for( int i = 0; i < 10; ++i )
		{
			File f;
			f.SetName( ssprintf( "%c", 'a' + i ) );
			f.dir = true;
			f.size = 0;
			f.hash = 0;

			fs.files.insert(f);
		}
	}
};

void sanity()
{
	TestFilenameDB db;
	for( int i = 0; i < 2; ++i )
	{
		vector<CString> list;
		db.GetDirListing( "a/*", list, true, true );
		ASSERT( list.size() == 10 );
		list.clear();

		db.GetDirListing( "a/*", list, false, true );
		ASSERT( list.size() == 110 );
		list.clear();

		db.FlushDirCache();
	}
	
}

/* This test continually reads directories and occasionally flushes. */
void run1( FilenameDB &db )
{
	RageTimer tm, flush;
	float flush_after_every = RandomFloat( 0.01, 0.05 );
	while( 1 ) //tm.PeekDeltaTime() < 2.0f )
	{
		if( flush.PeekDeltaTime() >= flush_after_every )
		{
			flush.GetDeltaTime();
			db.FlushDirCache();
			continue;
		}


		vector<CString> list;

		char levels = RandomInt( 1, 3 );
		CString dir;
		for( int i = 0; i < levels; ++i )
			dir += ssprintf( "%c/", '0' + RandomInt(0,9) );

		db.GetDirListing( dir, list, true, true );
	}
}

int run1_thread( void *p )
{
	printf("%i\n", (int) pthread_self());
	TestFilenameDB *db = (TestFilenameDB *) p;
	run1( *db );
	return 0;
}

void go()
{
	sanity();

	{
//		TestFilenameDB db;
//		run1( db );
	}
	
	/* Run the test in a few threads at once. */
	{
		const int threads = 10;
		RageThread t[threads];
		TestFilenameDB db;
		for( int i = 0; i < threads; ++i )
		{
			t[i].SetName( ssprintf("Test thread %i", i) );
			t[i].Create( run1_thread, &db );
		}

		for( int i = 0; i < threads; ++i )
			t[i].Wait();
	}
}

int main( int argc, char *argv[] )
{
	test_handle_args( argc, argv );

	test_init();


	go();

	test_deinit();

	exit(0);
}

